###############################################################################
# KoXo Exchange Online
# 01/07/2022
###############################################################################
function connect_exo {
    Param([string] $MyName, $MyPassword)
    $Cred = get_cred $MyName $MyPassword
    Connect-ExchangeOnline -Credential $Cred -ErrorAction SilentlyContinue
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Connection was successful to Microsoft Exchange Online Management"
        log_message "OK" $S
        Return $true
    } 
    else {
        $global:total_errors_count++
        $S = "Connection was unsuccessful to Microsoft Exchange Online Management || $($error[0])"
        log_message "ERROR" $S
        # Display error
        Write-host $S -foregroundcolor "red"
        Write-host "Press 'Enter' to continue ..."
        Read-Host   
        Return $false
    }   
}
###############################################################################
function user_exists {
    Param([string] $Email)
    $error.Clear()
    $User = Get-User -Identity $Email -ErrorAction SilentlyContinue
    return (![string]::IsNullOrEmpty($User))
}
###############################################################################
function user_info {
    Param([string] $identity)
    $Error.Clear()
    $User = Get-User -Identity $identity -ErrorAction SilentlyContinue
    return $User
}
###############################################################################
function usermailbox_info {
    Param([string] $identity)
    $Error.Clear()
    $User = Get-EXOMailbox -Identity $identity -ErrorAction SilentlyContinue
    return $User
}
###############################################################################
function usermailbox_stat {
    Param([string] $identity)
    $Error.Clear()
    $User = Get-MailboxStatistics -Identity $identity -ErrorAction SilentlyContinue
    return $User
}
###############################################################################
function set_regional_configuration {
    Param($UserDisplayName, $UserEmail, $Language, $TimeZone)
    $error.Clear()
    Set-MailboxRegionalConfiguration -Identity $UserEmail -Language $Language -TimeZone $TimeZone -LocalizeDefaultFolderName:$true
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully configure language ""$($Language)"" and timezone ""$($TimeZone)"" for ""$($UserDisplayName)"" ($($UserEmail))"
        log_message "OK" $S
    }
    else {
        $global:warnings_count++
        $S = "Configuring language ""$($Language)"" and timezone ""$($TimeZone)"" for ""$($UserDisplayName)"" ($($UserEmail)). Error = $($error[0])"
        log_message "WARNING" $S
    }
}
###############################################################################
function get_deleted_mailbox {
    $Error.Clear()
    $Users = Get-Mailbox -SoftDeletedMailbox
    return $Users
}
###############################################################################
function group_info {
    Param([string] $grp)
    $Error.Clear()
    $Group = Get-Group -Identity $grp -ErrorAction SilentlyContinue
    return $Group
}
###############################################################################
function distributiongroupmember_info {
    Param([string] $grp)
    $Error.Clear()
    $Group = Get-DistributionGroupMember -Identity $grp -ResultSize Unlimited -ErrorAction SilentlyContinue
    $Group = $Group | Sort-Object -Property DisplayName
    $Members = New-Object System.Collections.ArrayList
    $Index = 0
    ForEach ($Member In $Group) {
        $Line = "{0,-5}{1,-40} {2}" -f ($Index + 1), $Member.DisplayName, $Member.WindowsLiveID
        $Index = $Members.Add($Line)
    }
    return $Members
}
###############################################################################
function create_user {
    Param($UserEmail,
        $UserPassword,
        $UserMustChangePassword,
        $UserHiddenFromAddressLists,
        $UserDisplayName,
        $UserFunction, $UserFirstName, $UserLastName,
        $UserInitials,
        $UserPhone, $UserMobilePhone, $UserFaxNumber,
        $UserWebPage,
        $UserDepartment, $UserOffice, $UserCompany,
        $UserEmail1, $UserEmail2, $UserEmail3, $UserEmail4, $UserEmail5,
        $UserNotes,
        $UserPicture,
        $MemberOf,
        $CustomAttribute1, $CustomAttribute2, $CustomAttribute3, $CustomAttribute4,
        $CustomAttribute5, $CustomAttribute6, $CustomAttribute7, $CustomAttribute8,
        $CustomAttribute9, $CustomAttribute10, $CustomAttribute11, $CustomAttribute12,
        $CustomAttribute13, $CustomAttribute14, $CustomAttribute15)
    $UserIdentity = $UserEmail.Split("@")[0]
    if ($UserMustChangePassword -isnot [Boolean]) { $UserMustChangePassword = $false }
    if ($UserHiddenFromAddressLists -isnot [Boolean]) { $UserHiddenFromAddressLists = $false }
    $UserNotes = $UserNotes -replace ("/", "`n")
    $error.Clear()
    # Secure the password
    if (![String]::IsNullorEmpty($UserPassword)) {
        $SecuredPassword = ConvertTo-SecureString $UserPassword -AsPlainText -Force
    }
    else {
        $SecuredPassword = $UserPassword
    }
    New-Mailbox -Name $UserIdentity -Password $SecuredPassword -MicrosoftOnlineServicesID $UserEmail -ResetPasswordOnNextLogon $UserMustChangePassword -FirstName $UserFirstName -LastName $UserLastName -DisplayName $UserDisplayName -Initials $UserInitials
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully create mailbox for $($UserDisplayName) ($($UserEmail))"
        log_message "OK" $S
    }
    else {
        $global:errors_count++
        $S = "Creating mailbox for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])"
        log_message "ERROR" $S
    }
    # Build the list of Proxy Email Addresses
    $EmailAddresses = New-Object System.Collections.ArrayList
    if (![String]::IsNullorEmpty($UserEmail)) {
        $index = $EmailAddresses.Add("SIP:$($UserEmail)")
        $index = $EmailAddresses.Add("SMTP:$($UserEmail)")
        if (![String]::IsNullorEmpty($UserEmail1)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail1)")
        }
        if (![String]::IsNullorEmpty($UserEmail2)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail2)")
        }
        if (![String]::IsNullorEmpty($UserEmail3)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail3)")
        }
        if (![String]::IsNullorEmpty($UserEmail4)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail4)")
        }
        if (![String]::IsNullorEmpty($UserEmail5)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail5)")
        }
    }
    # Add Proxy Email Addresses and Custom Attributes    
    $error.Clear()
    Set-Mailbox -Identity $UserIdentity -EmailAddresses $EmailAddresses -CustomAttribute1 $CustomAttribute1 -CustomAttribute2 $CustomAttribute2 -CustomAttribute3 $CustomAttribute3 -CustomAttribute4 $CustomAttribute4 -CustomAttribute5 $CustomAttribute5 -CustomAttribute6 $CustomAttribute6 -CustomAttribute7 $CustomAttribute7 -CustomAttribute8 $CustomAttribute8 -CustomAttribute9 $CustomAttribute9 -CustomAttribute10 $CustomAttribute10 -CustomAttribute11 $CustomAttribute11 -CustomAttribute12 $CustomAttribute12 -CustomAttribute13 $CustomAttribute13 -CustomAttribute14 $CustomAttribute14 -CustomAttribute15 $CustomAttribute15 -HiddenFromAddressListsEnabled $UserHiddenFromAddressLists
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully update mailbox for $($UserDisplayName) ($($UserEmail))"
        log_message "OK" $S
    }
    else {
        $global:errors_count++
        $S = "Updating mailbox for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])"
        log_message "ERROR" $S
    }    
    # Add User Misc. Attributes    
    $error.Clear()
    Set-User -Identity $UserIdentity -Title $UserFunction -Phone $UserPhone -MobilePhone $UserMobilePhone -Fax $UserFaxNumber -WebPage $UserWebPage -Department $UserDepartment -Office $UserOffice -Company $UserCompany -Notes $UserNotes
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully update user's informations for $($UserDisplayName) ($($UserEmail))"
        log_message "OK" $S
    }
    else {
        $global:errors_count++
        $S = "Updating user's informatios for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])"
        log_message "ERROR" $S
    }
    # Add membership to distribution groups
    $Groups = $MemberOf.Split(",")
    ForEach ($Group in $Groups) {
        if (![String]::IsNullorEmpty($Group)) {
            $error.Clear()
            Add-DistributionGroupMember -Identity $Group -Member $UserIdentity
            if ([String]::IsNullorEmpty($error[0])) {
                $S = "Successfully add distribution group ""$($Group)"" for $($UserDisplayName) ($($UserEmail))"
                write-host $S -ForegroundColor Green
                log_message "OK" $S
            }
            else {
              $global:errors_count++
              $S = "Error while adding distribution group ""$($Group)"" for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])"
              write-host $S -ForegroundColor Red
              log_message "ERROR" $S
            }				
        }
    }  
    # Add user's picture or photo / must be smaller than 10KB
    if (![String]::IsNullorEmpty($UserPicture)) {
        If ((Test-Path $UserPicture) -eq $True) {
            $error.Clear()
            Set-UserPhoto -Identity $UserEmail -PictureData $UserPicture
            if ([String]::IsNullorEmpty($error[0])) {
                $FileSize = (Get-Item $UserPicture).Length / 1024
                $S = "Successfully add user's picture ""$($UserPicture)"" ({0:n1}KB) for $($UserDisplayName) ($($UserEmail))" -f $FileSize
                log_message "OK" $S
            }
            else {
                $global:errors_count++
                $FileSize = (Get-Item $UserPicture).Length / 1024
                $S = "Adding user's picture ""$($UserPicture)"" ({0:n1}KB) for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])" -f $FileSize
                log_message "ERROR" $S
            }
        }   
    }
}
###############################################################################
function modify_user {
    Param($UserEmail,
        $UserMustChangePassword,
        $UserHiddenFromAddressLists,
        $UserDisplayName,
        $UserFunction, $UserFirstName, $UserLastName,
        $UserInitials,
        $UserPhone, $UserMobilePhone, $UserFaxNumber,
        $UserWebPage,
        $UserDepartment, $UserOffice, $UserCompany,
        $UserEmail1, $UserEmail2, $UserEmail3, $UserEmail4, $UserEmail5,
        $UserNotes,
        $UserPicture,
        $MemberOf,
        $CustomAttribute1, $CustomAttribute2, $CustomAttribute3, $CustomAttribute4,
        $CustomAttribute5, $CustomAttribute6, $CustomAttribute7, $CustomAttribute8,
        $CustomAttribute9, $CustomAttribute10, $CustomAttribute11, $CustomAttribute12,
        $CustomAttribute13, $CustomAttribute14, $CustomAttribute15
    )
    $UserIdentity = $UserEmail.Split("@")[0]   
    if ($UserMustChangePassword -isnot [Boolean]) { $UserMustChangePassword = $false }
    if ($UserHiddenFromAddressLists -isnot [Boolean]) { $UserHiddenFromAddressLists = $false }
    $UserNotes = $UserNotes -replace ("/", "`n")
    $error.Clear()
    # Remove added emails
    $Id = Get-EXOMailbox -Identity $UserEmail
    ForEach ($Email in $Id.EmailAddresses) {
        If (($Email -clike "smtp:*") -Or ($Email -clike "SIP:*")) {
            Set-Mailbox -Identity $UserEmail -EmailAddresses @{remove = $Email }
        }
    }	
    # Build the list of Proxy Email Addresses
    $EmailAddresses = New-Object System.Collections.ArrayList
    if (![String]::IsNullorEmpty($UserEmail)) {
        $index = $EmailAddresses.Add("SIP:$($UserEmail)")
        $index = $EmailAddresses.Add("SMTP:$($UserEmail)")
        if (![String]::IsNullorEmpty($UserEmail1)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail1)")
        }
        if (![String]::IsNullorEmpty($UserEmail2)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail2)")
        }
        if (![String]::IsNullorEmpty($UserEmail3)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail3)")
        }
        if (![String]::IsNullorEmpty($UserEmail4)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail4)")
        }
        if (![String]::IsNullorEmpty($UserEmail5)) {
            $index = $EmailAddresses.Add("smtp:$($UserEmail5)")
        }
    }
    # Add Proxy Email Addresses and Custom Attributes    
    $error.Clear()
    Set-Mailbox -Identity $UserEmail -EmailAddresses $EmailAddresses -CustomAttribute1 $CustomAttribute1 -CustomAttribute2 $CustomAttribute2 -CustomAttribute3 $CustomAttribute3 -CustomAttribute4 $CustomAttribute4 -CustomAttribute5 $CustomAttribute5 -CustomAttribute6 $CustomAttribute6 -CustomAttribute7 $CustomAttribute7 -CustomAttribute8 $CustomAttribute8 -CustomAttribute9 $CustomAttribute9 -CustomAttribute10 $CustomAttribute10 -CustomAttribute11 $CustomAttribute11 -CustomAttribute12 $CustomAttribute12 -CustomAttribute13 $CustomAttribute13 -CustomAttribute14 $CustomAttribute14 -CustomAttribute15 $CustomAttribute15 -HiddenFromAddressListsEnabled $UserHiddenFromAddressLists
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully update mailbox for $($UserDisplayName) ($($UserEmail))"
        write-host $S -ForegroundColor Green
        log_message "OK" $S
    }
    else {
        $global:errors_count++
        $S = "Updating mailbox for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])"
        write-host $S  -ForegroundColor Red
        log_message "ERROR" $S
    }
    # Update User Misc. Attributes 1    
    $error.Clear()
    Set-MsolUser -UserPrincipalName $UserEmail -Title $UserFunction -DisplayName $UserDisplayName -FirstName $UserFirstName -LastName $UserLastName -PhoneNumber $UserPhone -MobilePhone $UserMobilePhone -Fax $UserFaxNumber -Department $UserDepartment -Office $UserOffice -ErrorAction SilentlyContinue
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully update user's AzureAD informations for $($UserDisplayName) ($($UserEmail))"
        write-host $S -ForegroundColor Green
        log_message "OK" $S
    }
    else {
        $global:errors_count++
        $S = "Updating user's AzureAD informations for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])"
        write-host $S  -ForegroundColor Red
        log_message "ERROR" $S
    }
    # Update User Misc. Attributes 2   
    $error.Clear()
    Set-User -Identity $UserEmail -Initials $UserInitials -WebPage $UserWebPage -Company $UserCompany -Notes $UserNotes -ResetPasswordOnNextLogon $UserMustChangePassword 
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully update user's Exchange informations for $($UserDisplayName) ($($UserEmail))"
        write-host $S  -ForegroundColor Green
        log_message "OK" $S
    }
    else {
        $global:errors_count++
        $S = "Updating user's Exchange informations for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])"
        write-host $S  -ForegroundColor Red
        log_message "ERROR" $S
    }
    # Display distribution groups
    $DistributionGroups = Get-DistributionGroup -Filter "Members -like '$($Id.DistinguishedName)'" 
    if ($DistributionGroups -ne $null) {
        write-Host "Present distribution groups : " -ForegroundColor Cyan
        ForEach ($DistributionGroup In $DistributionGroups) {
            write-host "Name : ""$($DistributionGroup.Name)"", Email address : $($DistributionGroup.PrimarySmtpAddress), Alias : ""$($DistributionGroup.Alias)""" -ForegroundColor Cyan
        }    
    }
    ### Synchronize membership to distribution groups
    $Groups = $MemberOf.Split(",", [System.StringSplitOptions]::RemoveEmptyEntries);
    # Remove unwanted distribution groups
    if ($DistributionGroups -ne $null) {
        ForEach ($DistributionGroup in $DistributionGroups) {
            $index = $Groups.IndexOf($DistributionGroup.Alias)
            if ($index -eq -1) {
                $error.Clear()
                Remove-DistributionGroupMember -Identity $DistributionGroup.Alias -Member $UserIdentity -Confirm:$false
                if ([String]::IsNullorEmpty($error[0])) {
                    $S = "Successfully remove distribution group ""$($DistributionGroup.Name)"", Email address : ""$($DistributionGroup.PrimarySmtpAddress)"", Alias = ""$($DistributionGroup.Alias)"" for $($UserDisplayName) ($($UserEmail))"
                    write-host $S -ForegroundColor Green
                    log_message "OK" $S
                }
            }
        }
        $DistributionGroups = Get-DistributionGroup -Filter "Members -like '$($Id.DistinguishedName)'" 
    }
    # Add not present groups
    ForEach ($Group in $Groups) {
        $error.Clear()
        if (($DistributionGroups | Where-Object { $_.Alias -eq $Group }) -eq $null) {
            Add-DistributionGroupMember -Identity $Group -Member $UserIdentity
            if ([String]::IsNullorEmpty($error[0])) {
                $S = "Successfully add distribution group ""$($Group)"" for $($UserDisplayName) ($($UserEmail))"
                write-host $S -ForegroundColor Green
                log_message "OK" $S
            }
            else {
              $global:errors_count++
              $S = "Error while adding distribution group ""$($Group)"" for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])"
              write-host $S -ForegroundColor Red
              log_message "ERROR" $S
            }			
        }
    }  
    #
    # Add membership to security groups
    #
    ForEach ($Group in $Groups) {
        $error.Clear()
        $SecurityGroup = get_koxo_security_group $Group
        if ($SecurityGroup -ne $empty) {	 
            $User = Get-MsolUser -UserPrincipalName $UserEmail
            Add-MsolGroupMember -GroupObjectId $SecurityGroup.ObjectId -GroupMemberType "User" -GroupMemberObjectId $User.ObjectId
            if ([String]::IsNullorEmpty($error[0])) {
                $S = "Successfully add security group ""$($SecurityGroup.DisplayName)"" for $($UserDisplayName) ($($UserEmail))"
                Write-Host $S
                log_message "OK" $S
            }
        }   
    }  
    # Add user's picture or photo / must be smaller than 10KB
    if (![String]::IsNullorEmpty($UserPicture)) {
        If ((Test-Path $UserPicture) -eq $True) {
            $error.Clear()
            Set-UserPhoto -Identity $UserEmail -PictureData ([System.IO.File]::ReadAllBytes($UserPicture)) -Confirm:$false
            if ([String]::IsNullorEmpty($error[0])) {
                $FileSize = (Get-Item $UserPicture).Length / 1024
                $S = "Successfully update user's picture ""$($UserPicture)"" ({0:n1}KB) for $($UserDisplayName) ($($UserEmail))" -f $FileSize
                log_message "OK" $S
            }
            else {
                $global:errors_count++
                $FileSize = (Get-Item $UserPicture).Length / 1024
                $S = "Updating user's picture ""$($UserPicture)"" ({0:n1}KB) for $($UserDisplayName) ($($UserEmail)). Error = $($error[0])" -f $FileSize
                log_message "ERROR" $S
            }
        }   
    }
}
###############################################################################
function distribution_group_exists {
    Param([string] $GroupId)
    $error.Clear()
    $Group = Get-DistributionGroup -Identity $GroupId -ErrorAction SilentlyContinue
    return (![string]::IsNullOrEmpty($Group))
}
###############################################################################
function distributiongroup_info {
    Param([string] $grp)
    $Error.Clear()
    $Group = Get-DistributionGroup -Identity $grp -ErrorAction SilentlyContinue
    return $Group
}
###############################################################################
function distributiongroupmember_info {
    Param([string] $grp)
    $Error.Clear()
    $Group = Get-DistributionGroupMember -Identity $grp -ResultSize Unlimited -ErrorAction SilentlyContinue
    $Group = $Group | Sort-Object -Property DisplayName
    $Members = New-Object System.Collections.ArrayList
    $Index = 0
    ForEach ($Member In $Group) {
        $Line = "{0,-5}{1,-40} {2}" -f ($Index + 1), $Member.DisplayName, $Member.WindowsLiveID
        $Index = $Members.Add($Line)
    }
    return $Members
}
###############################################################################
function create_distribution_group {
    Param($GroupName, 
        $GroupDisplayName,
        $GroupAlias,
        $GroupEmail,
        $RequireSenderAuthenticationEnabled,
        $GroupMembers)
    $error.Clear()
    New-DistributionGroup -Name $GroupName -Alias $GroupAlias -PrimarySmtpAddress $GroupEmail -DisplayName $GroupDisplayName -MemberJoinRestriction Closed -MemberDepartRestriction Closed
    if ([String]::IsNullorEmpty($error[0])) {
        $error.Clear()
        Set-DistributionGroup -Identity $GroupEmail -RequireSenderAuthenticationEnabled $RequireSenderAuthenticationEnabled 
        $S = "Successfully create distribution group, Name=""$($GroupName)"" DisplayName=""$($GroupDisplayName)"", Email=""$($GroupEmail)"", Alias=""$($GroupAlias)"""
        log_message "OK" $S
        # Members
        $Members = $GroupMembers.Split(",", [System.StringSplitOptions]::RemoveEmptyEntries)
        ForEach ($Member in $Members) {
            if (![String]::IsNullorEmpty($Member)) {                
                $error.Clear()
                Add-DistributionGroupMember -Identity $GroupEmail -Member $Member
                if ([String]::IsNullorEmpty($error[0])) {
                    $S = "Successfully add member ""$($Member)"" for ""$($GroupName)"" ($($GroupEmail))"
                    log_message "OK" $S
                    Write-host $S -foregroundcolor "Green"
                }
                else {
                    $global:errors_count++
                    $S = "Adding member ""$($Member)"" for ""$($GroupName)"" ($($GroupEmail)). Error = $($error[0])"
                    log_message "ERROR" $S                  
                    Write-Host $S -Foregroundcolor "Red"
                }
            }   
        }
    }
    else {
        $global:errors_count++
        $S = "Creating distribution group, Name=""$($GroupName)"" DisplayName=""$($GroupDisplayName)"", Email=""$($GroupEmail)"", Alias=""$($GroupAlias)"". Error = $($error[0])"
        log_message "ERROR" $S
    }
}
###############################################################################
function modify_distribution_group {
    Param($GroupName, 
        $GroupDisplayName,
        $GroupAlias,
        $GroupEmail,
        $RequireSenderAuthenticationEnabled,
        $GroupMembers)
    $error.Clear()
    Set-DistributionGroup -Identity $GroupEmail -Alias $GroupAlias -DisplayName $GroupDisplayName -RequireSenderAuthenticationEnabled $RequireSenderAuthenticationEnabled -MemberJoinRestriction Closed -MemberDepartRestriction Closed
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully update distribution group, Name=""$($GroupName)"" DisplayName=""$($GroupDisplayName)"", Email=""$($GroupEmail)"", Alias=""$($GroupAlias)"""
        log_message "OK" $S
        $DistributionGroupMembers = Get-DistributionGroupMember -Identity $GroupEmail -ResultSize Unlimited
        $MembersDistrib = New-Object System.Collections.ArrayList
        ForEach ($Member in $DistributionGroupMembers) { $Temp = $MembersDistrib.Add($Member.WindowsLiveID) }
        $DesiredMembers = $GroupMembers.Split(",", [System.StringSplitOptions]::RemoveEmptyEntries)
        $MembersToAdd = New-Object System.Collections.ArrayList
        $MembersToRemove = New-Object System.Collections.ArrayList
        ForEach ($MemberDistrib in $MembersDistrib) {
            $index = $MembersToRemove.Add($MemberDistrib)
        }
        ForEach ($DesiredMember in $DesiredMembers) {
            $index = $MembersToAdd.Add($DesiredMember)
        }
        ForEach ($MemberDistrib in $MembersDistrib) {
            $index = $MembersToAdd.IndexOf($MemberDistrib)
            if ($index -gt -1) { $MembersToAdd.Remove($MemberDistrib) }
        }
        ForEach ($MyMember in $MembersDistrib) {
            $index = $DesiredMembers.IndexOf($MyMember)
            if ($index -gt -1) { $MembersToRemove.Remove($MyMember) }
        }
        ForEach ($Member in $MembersToRemove) {
            Write-Host "Removing member : $($Member)" -ForegroundColor Yellow
            $error.Clear()
            Remove-DistributionGroupMember -Identity $GroupEmail -Member $Member -Confirm:$false
            if ([String]::IsNullorEmpty($error[0])) {
                $S = "Successfully remove member ""$($Member)"" from distribution group : DisplayName=""$($GroupDisplayName)"", Name=""$($GroupName)"""
                log_message "OK" $S
                Write-host $S -foregroundcolor "Green"
            }
            else {
                $global:errors_count++
                $S = "Removing member ""$($Member)"" from distribution group : DisplayName=""$($GroupDisplayName)"", Name=""$($GroupName)"". Error = $($error[0])"
                log_message "ERROR" $S
                Write-Host $S -Foregroundcolor "Red"
            }
        }
        ForEach ($Member in $MembersToAdd) {
            Write-Host "Adding member : $($Member)" -ForegroundColor Yellow
            $error.Clear()
            Add-DistributionGroupMember -Identity $GroupEmail -Member $Member -Confirm:$false
            if ([String]::IsNullorEmpty($error[0])) {
                $S = "Successfully add member ""$($Member)"" to distribution group : DisplayName=""$($GroupDisplayName)"", Name=""$($GroupName)"""
                log_message "OK" $S
                Write-host $S -foregroundcolor "Green"
            }
            else {
                $global:errors_count++
                $S = "Adding member ""$($Member)"" to distribution group : DisplayName=""$($GroupDisplayName)"", Name=""$($GroupName)"". Error = $($error[0])"
                log_message "ERROR" $S
                Write-Host $S -Foregroundcolor "Red"
            }
        }
    }
    else {
        $global:errors_count++
        $S = "Updating distribution group, DisplayName=""$($GroupDisplayName)"", Name=""$($GroupName)"", Email=""$($GroupEmail)"", Alias=""$($GroupAlias)"". Error = $($error[0])"
        log_message "ERROR" $S
        Write-Host $S -Foregroundcolor "Red"
    }
}
###############################################################################
function delete_distribution_group {
    Param($GroupName,
        $GroupDisplayName,
        $GroupAlias,
        $GroupEmail)
    $error.Clear()
    Remove-DistributionGroup -Identity $GroupEmail -Confirm:$false
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully delete distribution group, Name=""$($GroupName)"", DisplayName=""$($GroupDisplayName)"", Email=""$($GroupEmail)"", Alias=""$($GroupAlias)"""
        log_message "OK" $S
    }
    else {
        $global:errors_count++
        $S = "Deleting distribution group, Name=""$($GroupName)"", DisplayName=""$($GroupDisplayName)"", Email=""$($GroupEmail)"", Alias=""$($GroupAlias)"". Error = $($error[0])"
        log_message "ERROR" $S
    }
}
###############################################################################
function dynamic_group_exists {
    Param([string] $grp)
    $error.Clear()
    $Group = Get-DynamicDistributionGroup -Identity $grp -ErrorAction SilentlyContinue
    return (![string]::IsNullOrEmpty($Group))
}
###############################################################################
function dynamicgroup_info {
    Param([string] $grp)
    $Error.Clear()
    $Group = Get-DynamicDistributionGroup -Identity $grp -ErrorAction SilentlyContinue
    return $Group
}
###############################################################################
function create_dynamic_distribution_group {
    Param($GroupName, 
        $GroupDisplayName,
        $GroupAlias,
        $GroupEmail,
        $RequireSenderAuthenticationEnabled,
        $GroupCustomAttribute)
    $error.Clear()
    New-DynamicDistributionGroup -Name $GroupName -IncludedRecipients MailboxUsers -Alias $GroupAlias -PrimarySmtpAddress $GroupEmail -DisplayName $GroupDisplayName -ConditionalCustomAttribute1 $GroupCustomAttribute
    if ([String]::IsNullorEmpty($error[0])) {
        Set-DynamicDistributionGroup -Identity $GroupName -RequireSenderAuthenticationEnabled $RequireSenderAuthenticationEnabled
        $S = "Successfully create dynamic distribution group, Name=""$($GroupName)"" DisplayName=""$($GroupDisplayName)"", Email=""$($GroupEmail)"" ,Alias=""$($GroupAlias)"""
        log_message "OK" $S
    }
    else {
        $global:errors_count++
        $S = "Creating dynamic distribution group, Name=""$($GroupName)"" DisplayName=""$($GroupDisplayName)"", Email=""$($GroupEmail)"", Alias=""$($GroupAlias)"". Error = $($error[0])"
        log_message "ERROR" $S
    }
}
###############################################################################
function modify_dynamic_distribution_group {
    Param($GroupName,
        $GroupAlias,
        $GroupDisplayName,
        $RequireSenderAuthenticationEnabled,
        $GroupCustomAttribute,
        $GroupEmail)
    $error.Clear()
    Set-DynamicDistributionGroup -Identity $GroupEmail -IncludedRecipients MailboxUsers -Alias $GroupAlias -Name $GroupName -DisplayName $GroupDisplayName -RequireSenderAuthenticationEnabled $RequireSenderAuthenticationEnabled -ConditionalCustomAttribute1 $GroupCustomAttribute
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully update dynamic distribution group, Name=""$($GroupName)"", DisplayName=""$($GroupDisplayName)"", Email=""$($GroupEmail)"", Alias=""$($GroupAlias)"""
        log_message "OK" $S
        write-host $S
    }
    else {
        $global:errors_count++
        $S = "Updating dynamic distribution group, Name=""$($GroupName)"", DisplayName=""$($GroupDisplayName)"", Email=""$($GroupEmail)"", Alias=""$($GroupAlias)"". Error = $($error[0])"
        log_message "ERROR" $S
        write-host $S
    }
}
###############################################################################
function delete_dynamic_distribution_group {
    Param($GroupName,
        $GroupDisplayName,
        $GroupAlias,
        $GroupEmail)
    $error.Clear()
    Remove-DynamicDistributionGroup -Identity $GroupEmail -Confirm:$false
    if ([String]::IsNullorEmpty($error[0])) {
        $S = "Successfully delete dynamic distribution group, Name=""$($GroupName)"", DisplayName=""$($GroupDisplayName)"", Alias=""$($GroupAlias)"", Email=""$($GroupEmail)"""
        log_message "OK" $S
    }
    else {
        $global:errors_count++
        $S = "Deleting dynamic distribution group, Name=""$($GroupName)"", DisplayName=""$($GroupDisplayName)"", Alias=""$($GroupAlias)"", Email=""$($GroupEmail)"". Error = $($error[0])"
        log_message "ERROR" $S
    }
}
###############################################################################
function get_deleted_mailbox {
    $Error.Clear()
    $Users = Get-EXOMailbox -SoftDeletedMailbox
    return $Users
}
